from compgeom.pnt2d import Pnt2D
from compgeom.compgeom import CompGeom
from geometry.curves.curve import Curve
import math


class Polyline(Curve):
    def __init__(self, _pts=None):
        super(Curve, self).__init__()
        self.type = 'POLYLINE'
        self.pts = []
        if _pts is not None:
            for i in range(len(_pts)):
                self.pts.append(Pnt2D(_pts[i].getX(), _pts[i].getY()))
        self.nPts = len(self.pts)

    # ---------------------------------------------------------------------
    def isUnlimited(self):
        return True

    # ---------------------------------------------------------------------
    def addCtrlPoint(self, _x, _y):
        # **** COMPLETE HERE: POLYLINE_01 ****

    # ---------------------------------------------------------------------
    def isPossible(self):
        # **** COMPLETE HERE: POLYLINE_02 ****
        return True

    # ---------------------------------------------------------------------
    def getCtrlPoints(self):
        tempPts = []
        for i in range(0, self.nPts):
            tempPts.append(self.pts[i])
        return tempPts

    # ---------------------------------------------------------------------
    def setCtrlPoint(self, _id, _x, _y, _tol):
        # **** COMPLETE HERE: POLYLINE_03 ****
        return True

    # ---------------------------------------------------------------------
    def isStraight(self, _tol):
        if self.nPts < 2:
            return False
        for i in range(1, len(self.pts) - 1):
            if not CompGeom.pickLine(self.pts[0], self.pts[-1], self.pts[i], _tol):
                return False
        return True

    # ---------------------------------------------------------------------
    def isClosed(self):
        # **** COMPLETE HERE: POLYLINE_04 ****
        return False

    # ---------------------------------------------------------------------
    # Evaluate a point for a given parametric value.
    # Also return the segment of the evaluated point and
    # the local parametric value in segment
    def evalPointSeg(self, _t):
        if _t <= 0.0:
            return Pnt2D(self.pts[0].getX(), self.pts[0].getY()), 0, 0.0

        if _t >= 1.0:
            return Pnt2D(self.pts[-1].getX(), self.pts[-1].getY()), self.nPts - 1, 1.0

        length = self.length()
        s = _t * length
        loc_t = 1.0
        prev_id = 0
        next_id = 0
        lenToSeg = 0.0

        for i in range(1, len(self.pts)):
            prev_id = i - 1
            next_id = i
            dist = math.sqrt((self.pts[i].getX() - self.pts[i - 1].getX()) *
                             (self.pts[i].getX() - self.pts[i - 1].getX()) +
                             (self.pts[i].getY() - self.pts[i - 1].getY()) *
                             (self.pts[i].getY() - self.pts[i - 1].getY()))

            if (lenToSeg + dist) >= s:
                loc_t = (s - lenToSeg) / dist
                break

            lenToSeg += dist

        x = self.pts[prev_id].getX() + loc_t * \
            (self.pts[next_id].getX() - self.pts[prev_id].getX())
        y = self.pts[prev_id].getY() + loc_t * \
            (self.pts[next_id].getY() - self.pts[prev_id].getY())

        return Pnt2D(x, y), prev_id, loc_t

    # ---------------------------------------------------------------------
    def evalPoint(self, _t):
        # **** COMPLETE HERE: POLYLINE_05 ****
        return pt

    # ---------------------------------------------------------------------
    def evalPointTangent(self, _t):
        # **** COMPLETE HERE: POLYLINE_06 ****
        return pt, tangVec

    # ---------------------------------------------------------------------
    def splitRaw(self, _t):
        # **** COMPLETE HERE: POLYLINE_07 ****
        return left, right

    # ---------------------------------------------------------------------
    def split(self, _t):
        left, right = self.splitRaw(_t) 
        return left, right

    # ---------------------------------------------------------------------
    def join(self, _joinCurve, _pt, _tol):
        if (_joinCurve.getType() != 'LINE') and (_joinCurve.getType() != 'POLYLINE'):
            return False, None, 'Cannot join segments:\n A POLYLINE curve may be joined only with a LINE or a POLYLINE.'

        # Get the relative orientation of the two curves.
        # The first curve is always the self.
        # It is assumed that the given point _pt is the common point of
        # the curves to be joined.
        selfPtInit = self.getPntInit()
        selfPtEnd = self.getPntEnd()
        otherPtInit = _joinCurve.getPntInit()
        otherPtEnd = _joinCurve.getPntEnd()
        if ((Pnt2D.euclidiandistance(selfPtEnd, _pt) < _tol) and
            (Pnt2D.euclidiandistance(otherPtEnd, _pt) < _tol)):
            selfReversed = False
            otherReversed = True
        elif ((Pnt2D.euclidiandistance(selfPtInit, _pt) < _tol) and
              (Pnt2D.euclidiandistance(otherPtInit, _pt) < _tol)):
            selfReversed = True
            otherReversed = False
        elif ((Pnt2D.euclidiandistance(selfPtInit, _pt) < _tol) and
              (Pnt2D.euclidiandistance(otherPtEnd, _pt) < _tol)):
            selfReversed = True
            otherReversed = True
        else: # default: self is left and other is right
            selfReversed = False
            otherReversed = False

        # Create a polyline with an ordered list of points.
        pts = []
        if selfReversed:
            for i in range(len(self.pts) - 1, -1, -1):
                pts.append(self.pts[i])
        else:
            for i in range(0, len(self.pts)):
                pts.append(self.pts[i])
        
        otherPts = _joinCurve.getCtrlPoints()
        if otherReversed:
            for i in range(len(otherPts) - 2, -1, -1):
                pts.append(otherPts[i])
        else:
            for i in range(1, len(otherPts)):
                pts.append(otherPts[i])
        crv = Polyline(pts)

        return True, crv, None

    # ---------------------------------------------------------------------
    def getEquivPolyline(self):
        equivPoly = []
        if self.nPts < 2:
            return equivPoly
        for i in range(0, self.nPts):
            equivPoly.append(self.pts[i])
        return equivPoly

    # ---------------------------------------------------------------------
    def getEquivPolylineCollecting(self, _pt):
        # **** COMPLETE HERE: POLYLINE_08 ****

    # ---------------------------------------------------------------------
    # Returns the closest point on polyline for a given point.
    # Also returns the segment of the closest point and the arc
    # length from the curve init to the closest point.
    def closestPointSeg(self, _x, _y):
        pt = Pnt2D(_x, _y)
        d, clstPtSeg, t = CompGeom.getClosestPointSegment(self.pts[0], self.pts[1], pt)
        xOn = clstPtSeg.getX()
        yOn = clstPtSeg.getY()
        dmin = d
        seg = 0

        for i in range(1, self.nPts - 1):
            d, clstPtSeg, t = CompGeom.getClosestPointSegment(self.pts[i], self.pts[i + 1], pt)
            if d < dmin:
                xOn = clstPtSeg.getX()
                yOn = clstPtSeg.getY()
                dmin = d
                seg = i

        arcLen = 0.0
        for i in range(0, seg):
            arcLen += math.sqrt((self.pts[i + 1].getX() - self.pts[i].getX()) *
                                (self.pts[i + 1].getX() - self.pts[i].getX()) +
                                (self.pts[i + 1].getY() - self.pts[i].getY()) *
                                (self.pts[i + 1].getY() - self.pts[i].getY()))
        arcLen += math.sqrt((xOn - self.pts[seg].getX()) *
                            (xOn - self.pts[seg].getX()) +
                            (yOn - self.pts[seg].getY()) *
                            (yOn - self.pts[seg].getY()))

        clstPt = Pnt2D(xOn, yOn)
        return clstPt, dmin, seg, arcLen

    # ---------------------------------------------------------------------
    def closestPoint(self, _x, _y):
        # **** COMPLETE HERE: POLYLINE_09 ****
        return True, clstPt, dmin, t, tangVec

    # ---------------------------------------------------------------------
    def closestPointParam(self, _x, _y, _tStart):
        clstPt, tangVec = self.evalPointTangent(_tStart)
        if ((abs(clstPt.getX() - _x) < Curve.COORD_TOL) and
            (abs(clstPt.getY() - _y) < Curve.COORD_TOL)):
            return True, clstPt, 0.0, _tStart, tangVec

        status, clstPt, dist, t, tangVec = self.closestPoint(_x, _y)
        return status, clstPt, dist, t, tangVec

    # ---------------------------------------------------------------------
    def getBoundBox(self):
        # **** COMPLETE HERE: POLYLINE_10 ****
        return xmin, xmax, ymin, ymax

    # ---------------------------------------------------------------------
    def getXinit(self):
        return self.pts[0].getX()

    # ---------------------------------------------------------------------
    def getYinit(self):
        return self.pts[0].getY()

    # ---------------------------------------------------------------------
    def getXend(self):
        return self.pts[-1].getX()

    # ---------------------------------------------------------------------
    def getYend(self):
        return self.pts[-1].getY()

    # ---------------------------------------------------------------------
    def getPntInit(self):
        return self.pts[0]

    # ---------------------------------------------------------------------
    def getPntEnd(self):
        return self.pts[-1]

    # ---------------------------------------------------------------------
    def length(self):
        if self.nPts < 2:
            return 0.0
        L = 0.0
        for i in range(0, self.nPts - 1):
            L += math.sqrt((self.pts[i + 1].getX() - self.pts[i].getX()) *
                           (self.pts[i + 1].getX() - self.pts[i].getX()) +
                           (self.pts[i + 1].getY() - self.pts[i].getY()) *
                           (self.pts[i + 1].getY() - self.pts[i].getY()))
        return L
