from ntpath import join
from mesh.meshgenerator import MeshGenerator
from compgeom.pnt2d import Pnt2D


class TransfinTrilinear(MeshGenerator):

    def __init__(self):
        self.n = 0

    # ---------------------------------------------------------------------
    # Auxiliary method for indexing a triangular matrix with size = n+1
    # stored in a vector.
    def mtxIdsToVecId(self, _i, _j):
        return int(((_i * (2 * self.n - _i + 3)) / 2) + _j)

    # ---------------------------------------------------------------------
    def setLoops(self, _loops):
        self.n = 0

        # Check for just one loop
        if len(_loops) != 1:
            return False

        # Check for exactly three curves on loop
        segSbdvs = _loops[0]
        if len(segSbdvs) != 3:
            return False

        # Check for equal number of subdivisions on all segments of loop
        if segSbdvs[0] != segSbdvs[1]:
            return False
        if segSbdvs[0] != segSbdvs[2]:
            return False

        # Save parameters of valid boundary loop
        self.n = segSbdvs[0]

        return True

    # ---------------------------------------------------------------------
    def generateMesh(self, _bdryPnts):
        pts = []
        conn = []
        # Total number of boundary nodes must be consistent with
        # number of boundary segments
        if len(_bdryPnts) != self.n * 3:
            return False, pts, conn

        # Allocate point vector in row major order
        pt = Pnt2D(0.0, 0.0)
        for i in range(0, self.n + 1):
            for j in range(0, self.n - i + 1):
                pts.append(pt)

        # Input nodes from first boundary curve
        idBound = 0
        i = 0
        for j in range(0, self.n):
            k = self.mtxIdsToVecId(i, j)
            pts[k] = _bdryPnts[idBound]
            idBound += 1

        # Input node from second boundary curve
        for i in range(0, self.n):
            j = self.n - i
            # *** COMPLETE HERE - TRANSFINTRILINEAR: 01 ***

        # Input node from third boundary curve
        # *** COMPLETE HERE - TRANSFINTRILINEAR: 02 ***

        # Generate interior nodes
        k_i_0_j_0 = self.mtxIdsToVecId(0, 0)
        k_i_0_j_n = self.mtxIdsToVecId(0, self.n)
        k_i_n_j_0 = self.mtxIdsToVecId(self.n, 0)
        for i in range(1, self.n):
            k_i_j_n__i = self.mtxIdsToVecId(i, self.n - i)
            k_i_j_0 = self.mtxIdsToVecId(i, 0)
            for j in range(1, self.n - i):
                k = self.mtxIdsToVecId(i, j)
                k_i__j_j_0 = self.mtxIdsToVecId(i + j, 0)
                k_i_0_j__i = self.mtxIdsToVecId(0, i + j)
                k_i_0_j = self.mtxIdsToVecId(0, j)
                k_i_n__j_j = self.mtxIdsToVecId(self.n - j, j)
                pts[k] =                                                     \
                        (pts[k_i_j_n__i] * j / (self.n - i) +                \
                         pts[k_i_j_0]    * (self.n - i - j) / (self.n - i) + \
                         pts[k_i__j_j_0] * i / (i + j) +                     \
                         pts[k_i_0_j__i] * j / (i + j) +                     \
                         pts[k_i_0_j]    * (self.n - i - j) / (self.n - j) + \
                         pts[k_i_n__j_j] * i / (self.n - j) -                \
                         pts[k_i_0_j_0]  * (self.n - i - j) / (self.n) -     \
                         pts[k_i_0_j_n]  * j / self.n -                      \
                         pts[k_i_n_j_0]  * i / self.n) * 0.5
  
        # Generate triangular elements
        # *** COMPLETE HERE - TRANSFINTRILINEAR: 03 ***
            # Generate last triangle of row to complete patch
            n0 = self.mtxIdsToVecId(i    , self.n - i - 1)
            n1 = self.mtxIdsToVecId(i    , self.n - i)
            n2 = self.mtxIdsToVecId(i + 1, self.n - i - 1)
            tria = []
            tria.append(n0)
            tria.append(n1)
            tria.append(n2)
            conn.append(tria)
 
        return True, pts, conn
